@@ -20,23 +20,6 @@
 #ifndef GREP_GREP_H
 #define GREP_GREP_H 1
 
-/* Function pointer types.  */
-typedef void (*compile_fp_t) (char const *, size_t);
-typedef size_t (*execute_fp_t) (char const *, size_t, size_t *, char const *);
-
-/* grep.c expects the matchers vector to be terminated by an entry
-   with a NULL name, and to contain at least one entry. */
-struct matcher
-{
-  const char *name;
-  compile_fp_t compile;
-  execute_fp_t execute;
-};
-extern const struct matcher matchers[];
-
-extern const char before_options[];
-extern const char after_options[];
-
 /* The following flags are exported from grep for the matchers
    to look at. */
 extern int match_icase;		/* -i */
